This is an automated email from the ASF dual-hosted git repository.

reshke pushed a commit to branch main
in repository https://gitbox.apache.org/repos/asf/cloudberry.git

commit f011f6cf1002c8362d1532ed8d424e1f37bc4c50
Author: Alvaro Herrera <[email protected]>
AuthorDate: Wed Jul 27 07:55:13 2022 +0200

    Allow "in place" tablespaces.
    
    This is a backpatch to branches 10-14 of the following commits:
    
    7170f2159fb2 Allow "in place" tablespaces.
    c6f2f01611d4 Fix pg_basebackup with in-place tablespaces.
    f6f0db4d6240 Fix pg_tablespace_location() with in-place tablespaces
    7a7cd84893e0 doc: Remove mention to in-place tablespaces for 
pg_tablespace_location()
    5344723755bd Remove unnecessary Windows-specific basebackup code.
    
    In-place tablespaces were introduced as a testing helper mechanism, but
    they are going to be used for a bugfix in WAL replay to be backpatched
    to all stable branches.
    
    I (Álvaro) had to adjust some code to account for lack of
    get_dirent_type() in branches prior to 14.
    
    Author: Thomas Munro <[email protected]>
    Author: Michaël Paquier <[email protected]>
    Author: Álvaro Herrera <[email protected]>
    Discussion: 
https://postgr.es/m/[email protected]
---
 doc/src/sgml/config.sgml          | 19 +++++++++++++++++++
 src/backend/access/transam/xlog.c |  8 ++++++++
 src/backend/commands/tablespace.c | 39 ++++++++++++++++++++++++++++++++-------
 src/backend/utils/adt/misc.c      | 29 +++++++++++++++++++++++++++++
 src/backend/utils/misc/guc.c      | 12 ++++++++++++
 src/include/commands/tablespace.h |  2 ++
 6 files changed, 102 insertions(+), 7 deletions(-)

diff --git a/doc/src/sgml/config.sgml b/doc/src/sgml/config.sgml
index bd61286e042..bc3d0d1bd14 100644
--- a/doc/src/sgml/config.sgml
+++ b/doc/src/sgml/config.sgml
@@ -10458,6 +10458,25 @@ dynamic_library_path = 
'C:\tools\postgresql;H:\my_project\lib;$libdir'
     </para>
 
     <variablelist>
+     <varlistentry id="guc-allow-in-place-tablespaces" 
xreflabel="allow_in_place_tablespaces">
+      <term><varname>allow_in_place_tablespaces</varname> 
(<type>boolean</type>)
+      <indexterm>
+        <primary><varname>allow_in_place_tablespaces</varname> configuration 
parameter</primary>
+      </indexterm>
+      </term>
+      <listitem>
+       <para>
+        Allows tablespaces to be created as directories inside
+        <filename>pg_tblspc</filename>, when an empty location string
+        is provided to the <command>CREATE TABLESPACE</command> command.  This
+        is intended to allow testing replication scenarios where primary and
+        standby servers are running on the same machine.  Such directories
+        are likely to confuse backup tools that expect to find only symbolic
+        links in that location.  Only superusers can change this setting.
+       </para>
+      </listitem>
+     </varlistentry>
+
      <varlistentry id="guc-allow-system-table-mods" 
xreflabel="allow_system_table_mods">
       <term><varname>allow_system_table_mods</varname> (<type>boolean</type>)
       <indexterm>
diff --git a/src/backend/access/transam/xlog.c 
b/src/backend/access/transam/xlog.c
index cff69879aa1..07831e9b098 100644
--- a/src/backend/access/transam/xlog.c
+++ b/src/backend/access/transam/xlog.c
@@ -11918,6 +11918,14 @@ do_pg_start_backup(const char *backupidstr, bool fast, 
TimeLineID *starttli_p,
 
                        snprintf(fullpath, sizeof(fullpath), "pg_tblspc/%s", 
de->d_name);
 
+                       /*
+                        * Skip anything that isn't a symlink/junction.  For 
testing only,
+                        * we sometimes use allow_in_place_tablespaces to create
+                        * directories directly under pg_tblspc, which would 
fail below.
+                        */
+                       if (get_dirent_type(fullpath, de, false, ERROR) != 
PGFILETYPE_LNK)
+                               continue;
+
 #if defined(HAVE_READLINK) || defined(WIN32)
                        rllen = readlink(fullpath, linkpath, sizeof(linkpath));
                        if (rllen < 0)
diff --git a/src/backend/utils/adt/misc.c b/src/backend/utils/adt/misc.c
index eb8877fd7e8..eba029daa74 100644
--- a/src/backend/utils/adt/misc.c
+++ b/src/backend/utils/adt/misc.c
@@ -15,6 +15,7 @@
 #include "postgres.h"
 
 #include <sys/file.h>
+#include <sys/stat.h>
 #include <dirent.h>
 #include <fcntl.h>
 #include <math.h>
@@ -312,6 +313,9 @@ pg_tablespace_location(PG_FUNCTION_ARGS)
        char            sourcepath[MAXPGPATH];
        char            targetpath[MAXPGPATH];
        int                     rllen;
+#ifndef WIN32
+       struct stat st;
+#endif
 
        /*
         * It's useful to apply this function to pg_class.reltablespace, wherein
@@ -336,6 +340,31 @@ pg_tablespace_location(PG_FUNCTION_ARGS)
         */
        snprintf(sourcepath, sizeof(sourcepath), "pg_tblspc/%u", tablespaceOid);
 
+       /*
+        * Before reading the link, check if the source path is a link or a
+        * junction point.  Note that a directory is possible for a tablespace
+        * created with allow_in_place_tablespaces enabled.  If a directory is
+        * found, a relative path to the data directory is returned.
+        */
+#ifdef WIN32
+       if (!pgwin32_is_junction(sourcepath))
+               PG_RETURN_TEXT_P(cstring_to_text(sourcepath));
+#else
+       if (lstat(sourcepath, &st) < 0)
+       {
+               ereport(ERROR,
+                               (errcode_for_file_access(),
+                                errmsg("could not stat file \"%s\": %m",
+                                               sourcepath)));
+       }
+
+       if (!S_ISLNK(st.st_mode))
+               PG_RETURN_TEXT_P(cstring_to_text(sourcepath));
+#endif
+
+       /*
+        * In presence of a link or a junction point, return the path pointing 
to.
+        */
        rllen = readlink(sourcepath, targetpath, sizeof(targetpath));
        if (rllen < 0)
                ereport(ERROR,
diff --git a/src/backend/utils/misc/guc.c b/src/backend/utils/misc/guc.c
index 353c4988a0a..cb3b1a1cbdd 100644
--- a/src/backend/utils/misc/guc.c
+++ b/src/backend/utils/misc/guc.c
@@ -51,6 +51,7 @@
 #include "catalog/index.h"
 #include "commands/async.h"
 #include "commands/prepare.h"
+#include "commands/tablespace.h"
 #include "commands/trigger.h"
 #include "commands/user.h"
 #include "commands/vacuum.h"
@@ -2045,6 +2046,17 @@ static struct config_bool ConfigureNamesBool[] =
                NULL, NULL, NULL
        },
 
+       {
+               {"allow_in_place_tablespaces", PGC_SUSET, DEVELOPER_OPTIONS,
+                       gettext_noop("Allows tablespaces directly inside 
pg_tblspc, for testing."),
+                       NULL,
+                       GUC_NOT_IN_SAMPLE
+               },
+               &allow_in_place_tablespaces,
+               false,
+               NULL, NULL, NULL
+       },
+
        {
                {"lo_compat_privileges", PGC_SUSET, COMPAT_OPTIONS_PREVIOUS,
                        gettext_noop("Enables backward compatibility mode for 
privilege checks on large objects."),
diff --git a/src/include/commands/tablespace.h 
b/src/include/commands/tablespace.h
index 1f41964cf75..fe13c5d75d7 100644
--- a/src/include/commands/tablespace.h
+++ b/src/include/commands/tablespace.h
@@ -20,6 +20,8 @@
 #include "nodes/parsenodes.h"
 #include "storage/dbdirnode.h"
 
+extern bool allow_in_place_tablespaces;
+
 /* XLOG stuff */
 #define XLOG_TBLSPC_CREATE             0x00
 #define XLOG_TBLSPC_DROP               0x10


---------------------------------------------------------------------
To unsubscribe, e-mail: [email protected]
For additional commands, e-mail: [email protected]

Reply via email to
diff --git a/src/backend/commands/tablespace.c 
b/src/backend/commands/tablespace.c
index 9175ebfb5ba..3d7d040c462 100644
--- a/src/backend/commands/tablespace.c
+++ b/src/backend/commands/tablespace.c
@@ -113,6 +113,7 @@
 /* GUC variables */
 char      *default_tablespace = NULL;
 char      *temp_tablespaces = NULL;
+bool           allow_in_place_tablespaces = false;
 
 
 static void create_tablespace_directories(const char *location,
@@ -295,6 +296,7 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
        Datum           newOptions;
        List       *nonContentOptions = NIL;
        char       *fileHandler = NULL;
+       bool            in_place;
 
        /* Must be super user */
        if (!superuser())
@@ -362,12 +364,15 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
                                (errcode(ERRCODE_INVALID_NAME),
                                 errmsg("tablespace location cannot contain 
single quotes")));
 
+       in_place = allow_in_place_tablespaces && strlen(location) == 0;
+
        /*
         * Allowing relative paths seems risky
         *
         * this also helps us ensure that location is not empty or whitespace
         */
-       if (!is_absolute_path(location))
+       if (!in_place && !is_absolute_path(location))
                ereport(ERROR,
                                (errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
                                 errmsg("tablespace location must be an 
absolute path")));
@@ -862,20 +867,40 @@ create_tablespace_directories(const char *location, const 
Oid tablespaceoid)
        char       *location_with_dbid_dir;
        char       *location_with_version_dir;
        struct stat st;
+       bool            in_place;
 
        elog(DEBUG5, "creating tablespace directories for tablespaceoid %d on 
dbid %d",
                tablespaceoid, GpIdentity.dbid);
 
        linkloc = psprintf("pg_tblspc/%u", tablespaceoid);
+
+       /*
+        * If we're asked to make an 'in place' tablespace, create the directory
+        * directly where the symlink would normally go.  This is a 
developer-only
+        * option for now, to facilitate regression testing.
+        */
+       in_place = strlen(location) == 0;
+
+       if (in_place)
+       {
+               if (MakePGDirectory(linkloc) < 0 && errno != EEXIST)
+                       ereport(ERROR,
+                                       (errcode_for_file_access(),
+                                        errmsg("could not create directory 
\"%s\": %m",
+                                                       linkloc)));
+       }
+
        location_with_dbid_dir = psprintf("%s/%d", location, GpIdentity.dbid);
-       location_with_version_dir = psprintf("%s/%s", location_with_dbid_dir,
+       location_with_version_dir = psprintf("%s/%s", in_place ? linkloc : 
location_with_dbid_dir,
                                                                                
 GP_TABLESPACE_VERSION_DIRECTORY);
 
        /*
         * Attempt to coerce target directory to safe permissions.  If this 
fails,
         * it doesn't exist or has the wrong owner.
mode,
+        * because in that case we created the directory with the desired
+        * permissions.
         */
-       if (chmod(location, pg_dir_create_mode) != 0)
+       if (!in_place && chmod(location, pg_dir_create_mode) != 0)
        {
                if (errno == ENOENT)
                        ereport(ERROR,
@@ -949,13 +974,13 @@ create_tablespace_directories(const char *location, const 
Oid tablespaceoid)
        /*
         * In recovery, remove old symlink, in case it points to the wrong 
place.
         */
-       if (InRecovery)
+       if (!in_place && InRecovery)
                remove_tablespace_symlink(linkloc);
 
        /*
         * Create the symlink under PGDATA
         */
-       if (symlink(location_with_dbid_dir, linkloc) < 0)
+       if (!in_place && symlink(location_with_dbid_dir, linkloc) < 0)
                ereport(ERROR,
                                (errcode_for_file_access(),
                                 errmsg("could not create symbolic link \"%s\": 
@@ -362,12 +364,15 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
         * Allowing relative paths seems risky
         *
-        * this also helps us ensure that location is not empty or whitespace
+        * This also helps us ensure that location is not empty or whitespace,
+        * unless specifying a developer-only in-place tablespace.
         */
@@ -862,20 +867,40 @@ create_tablespace_directories(const char *location, const 
         * Attempt to coerce target directory to safe permissions.  If this 
fails,
-        * it doesn't exist or has the wrong owner.
+        * it doesn't exist or has the wrong owner.  Not needed for in-place 
mode,
%m",